/*
	file...: calc.c 
	purpose: calc shifts based on alignment and homologous shift files. 
*/
#include "defs.h"
#include <math.h>

/* static vars */

static PREDICT_REC pRec[MAX_RES_ATOM_NUM];
static int rix, seqIx;
static int sipRes, sisRes, scsRes;
static int aipRes, acpRes;
static int atomNum, atomIx;

/* procedure calls */
static void predict_sequence(), convert_atoms(), calc_pct_wt();
static void calc_amino_acid(), calc_atom(), calc_sdev(), shift_select();
static int calc_atom_shift(), flip_atom(), get_molmat_ix();

extern void print_ppm_title(), print_csi_title(), print_sequence();
extern void print_shifts_title();
extern void print_seq_names(), print_table(), calc_parm_weights();
extern void print_ppm(), print_shift(), print_csi_shift();
extern double calc_hom_ave_scor();
extern int sres_to_ix(), atom_to_index();

/*************************************************************/
/* perform orb algorithm 
*/
void do_orb()
{

	fprintf(stdout, "\nStarting calculations...\n\n");
	print_ppm_title(ppmfp);
	print_shifts_title(outfp);
	print_csi_title(csifp);
	print_sequence(verbfp, &seqArr[seqSelect], TRUE);
	print_seq_names(verbfp);
	predict_sequence();
	fprintf(stdout, "\nDone\n\n");

}	/* do_orb */


/*************************************************************/
/* process all shift predictions 
*/
static void predict_sequence()
{
	char scpRes;
	int sIdNum;


/* weights for local, global, etc homologies */
	calc_parm_weights();

/* convert all xx1, xx2 atoms to xx# if values the same */
	convert_atoms();
	fprintf(stdout, "\n");

/* process each amino acid in alignment*/
	for (rix=0; rix<alignLen; rix++) {
		scpRes = seqArr[seqSelect].residue[rix].pRes;
		sIdNum = seqArr[seqSelect].residue[rix].idNum;

/* find the valid amino acids */
		if (TRUE == AA_CVALID(scpRes)) {
			sipRes = AA_IX(scpRes);
			atomNum = aaTable[sipRes].atomNum;
			scsRes = seqArr[seqSelect].residue[rix].sRes;
			sisRes = sres_to_ix(scsRes);
			calc_amino_acid();
			fprintf(stdout, "Amino Acid = %c Structure = %c\n", scpRes, scsRes);

/* calc initial weights, then see which non stereo specific assigned
   shift fits best, then recalc weights */
			calc_pct_wt();
			shift_select();
			calc_pct_wt();

/* output results */
			calc_sdev();
			print_table(verbfp, pRec, rix, atomNum);
			print_ppm(ppmfp, pRec, scpRes, sIdNum, atomNum);
			print_shift(outfp, pRec, scpRes, sIdNum, atomNum);
			print_csi_shift(csifp, pRec, scpRes, sIdNum, atomNum);
		}
	}

}	/* predict_sequence */


/*************************************************************/
/* translate all XX1/XX2 atoms with same shift to XX*
*/
static void convert_atoms()
{
	register int i, ich, atomNum;
	ALIGN_RES_REC *arr;
	int len, px, newIx;
	char ch, *atom, cAtom[MAX_LINE_LEN];
	double s1, s2, c1;

	fprintf(stdout, " Convert all XX1/XX2 atoms with same shift to XX*...\n");
	for (seqIx=0; seqIx<numSeq; seqIx++) {
		for (rix=0; rix<alignLen; rix++) {
			arr = &seqArr[seqIx].residue[rix];
			if (TRUE == AA_CVALID(ch)) {
				ich = AA_IX(ch);
				atomNum = aaTable[ich].atomNum;
				for (i=0; i<atomNum; i++) {
					s1 = arr->shift1[i];
					s2 = arr->shift2[i];
					if ((s1 != NO_SHIFT_VAL) && (s2 == NO_SHIFT_VAL)) {
						atom = aaTable[ich].atom[i].rName;
						strcpy(cAtom, atom);
						if (flip_atom(cAtom) == TRUE) {
							px = atom_to_index(ich, cAtom);
							c1 = arr->shift1[px];
							if (s1 == c1) {
								fprintf(stdout, "  %s %c %s %s %f\n", 
									seqArr[seqIx].id, ch, atom, cAtom, c1);
								arr->shift1[i] = arr->shift1[px] = NO_SHIFT_VAL;
								len = strlen(cAtom) - 1;
								cAtom[len] = '#';
								newIx = atom_to_index(ich, cAtom);
								arr->shift1[newIx] = c1;
								arr->shift2[newIx] = NO_SHIFT_VAL;
							}
						}
					}
				}
			}
		}
	}

}	/* convert_atoms */


/*************************************************************/
/* create information table including homologous shift data 
*/
static void calc_amino_acid()
{
	PREDICT_REC *p;

/* fill in shifts from wishart tables at this time */
	for (atomIx=0; atomIx<atomNum; atomIx++) {
		p = &pRec[atomIx];
		p->atom = aaTable[sipRes].atom[atomIx].rName;
		p->dswVal = aaTable[sipRes].atom[atomIx].dswShift[sisRes];
		p->dswStd = aaTable[sipRes].atom[atomIx].dswStddev[sisRes];
		p->predict = NO_SHIFT_VAL;
	}

/* process the homologous sequence shifts for each atom */
	for (atomIx=0; atomIx<atomNum; atomIx++) {
		p = &pRec[atomIx];

/* only process atoms which have wishart values */
		if (p->dswVal != NO_SHIFT_VAL) {
			p->homNum = 0;

/* process each homologous sequence */
			for (seqIx=0; seqIx<numSeq; seqIx++) {

/* check for valid amino acids */
				acpRes = seqArr[seqIx].residue[rix].pRes;
				if ((seqIx != seqSelect) && (TRUE == AA_CVALID(acpRes))) {
					aipRes = AA_IX(acpRes);

/* process one shift value for one atom */
					calc_atom(p);
				}
			}
		}
	}

}	/* calc_amino_acid */


/*************************************************************/
/* calculate weights table for a single amino acid 
*/
static void calc_atom(p)
PREDICT_REC *p;
{
	int i, hAtomIx, mNum, numMatch;
	int match[MAX_RES_ATOM_NUM];
	MOL_MAT_REC *mp;
	register HOMOL_REC *h;
	double s1, s2;


/* find atoms that correspond with the atom to predict */
	mp = &molMat[sipRes][aipRes];
	mNum = molMatNum[sipRes][aipRes];
	numMatch = get_molmat_ix(mp->atom1, mNum, atomIx, match);

/* process each corresponding atom pair */
	for (i=0; i<numMatch; i++) {
		hAtomIx = mp->atom2[match[i]];

/* retrieve shift(s) if they were read in from shift file */
		s1 = seqArr[seqIx].residue[rix].shift1[hAtomIx];
		s2 = seqArr[seqIx].residue[rix].shift2[hAtomIx];

/* process shifts with values */
		if (s1 != NO_SHIFT_VAL) {

			h = &p->hom[p->homNum];
			h->seqId = seqArr[seqIx].id;
			h->atom = aaTable[aipRes].atom[hAtomIx].rName;
			h->molScor = mp->scor[match[i]];
			h->ppm[0] = s1; 
			h->ppm[1] = s2;
			h->selectShift = 0;

/* assign stereo specificity */
			h->ssaType = 0;
			if (h->atom[strlen(h->atom) - 1] == '#') 
				h->ssaType = 1;

/* case where two shifts are read */
			if (s2 != NO_SHIFT_VAL) {
				h->ssaType = 1;
				h->selectShift = -1;
			}

/* adjust shifts according to the amino acids we are comparing */
			if (TRUE == calc_atom_shift(p, h)) {

/* compute the weight we give this shift */
				h->wtScor = calc_hom_ave_scor(h->molScor,
					seqArr[seqIx].homGlobScor,
					seqArr[seqIx].residue[rix].homLocScor,
					seqArr[seqIx].residue[rix].homStructScor);
				h->wtScor *= seqArr[seqIx].userShiftBias;
				p->homNum++;
			}
		}
	}

}	/* calc_atom */


/*************************************************************/
/* adjust shift value according to the amino acids we are comparing
*/
static int calc_atom_shift(p, h)
PREDICT_REC *p;
HOMOL_REC *h;
{
	register int len, hix, aisRes;
	char cAtom[MAX_ATOM_NAME_LEN];
	double dswShift, dsw2, i1, i2, intrin;

/* find intrinsic shift value */
	aisRes = sres_to_ix(seqArr[seqIx].residue[rix].sRes);
	hix = atom_to_index(aipRes, h->atom);
	if (hix == ERROR_FLAG) {
		fprintf(stdout, "Unrecognized atom name: %s in %c\n", h->atom, acpRes);
		return(FALSE);
	}
	intrin = aaTable[aipRes].atom[hix].dswShift[aisRes];

/* if the selected sequence has unknown secondary structure, use
   homologous sequence structure */
	if (scsRes == CH_X)
		dswShift = aaTable[sipRes].atom[atomIx].dswShift[aisRes];
	else
		dswShift = p->dswVal;

/* perhaps we have a non stereo specific shift like HB# */
/* then find intrinsic shifts for HB1 and HB2 and take average */
	if ((intrin == NO_SHIFT_VAL) && (h->ssaType == 1)) {
		strcpy(cAtom, h->atom);
		len = strlen(cAtom) - 1;
		cAtom[len] = '1';
		hix = atom_to_index(aipRes, cAtom);
		if (hix != ERROR_FLAG) {
			i1 = aaTable[aipRes].atom[hix].dswShift[aisRes];
			cAtom[len] = '2';
			hix = atom_to_index(aipRes, cAtom);
			if (hix != ERROR_FLAG) {
				i2 = aaTable[aipRes].atom[hix].dswShift[aisRes];
				intrin = (i1 + i2) * 0.5;
				strcpy(cAtom, p->atom);
				if (flip_atom(cAtom) == TRUE) {
					hix = atom_to_index(sipRes, cAtom);
					dsw2 = aaTable[sipRes].atom[hix].dswShift[sisRes];
					dswShift = (dswShift + dsw2) * 0.5;
				}
			}
		}
	}

/* if not assigned by now, no sense continuing */
	if (intrin == NO_SHIFT_VAL) {
		fprintf(stdout, "No intrinsic shift for: %s in %c\n", h->atom, acpRes);
		return(FALSE);
	}

/* calculate adjusted shift value(s) */

	h->shift[0] = h->ppm[0] - intrin + dswShift;

	if (h->ppm[1] != NO_SHIFT_VAL) {
		h->shift[1] = h->ppm[1] - intrin + dswShift;
	}

	return(TRUE);

}	/* calc_atom_shift */


/*************************************************************/
/* create information table including homologous shift data 
*/
static void shift_select()
{
	register int i, ix, cAtomIx, num;
	register HOMOL_REC *h;
	PREDICT_REC *p;
	char cAtom[MAX_ATOM_NAME_LEN];
	double predict;

/* process each atom */
	for (ix=0; ix<atomNum; ix++) {
		p = &pRec[ix];
		num = p->homNum;
		predict = p->predict;

		for (i=0; i<num; i++) {
			h = &p->hom[i];
			if ((h->wtScor > 0) && (h->selectShift < 0)) {
				strcpy(cAtom, p->atom);
				if (TRUE == flip_atom(cAtom)) {
					cAtomIx = atom_to_index(sipRes, cAtom);
					if (predict < pRec[cAtomIx].predict)
						h->selectShift = 0;
					else
						h->selectShift = 1;
				} else {
					if (fabs(h->shift[0] - predict) <
						fabs(h->shift[1] - predict))
						h->selectShift = 0;
					else
						h->selectShift = 1;
				}
			}
		}
	}

}	/* shift_select */


/*************************************************************/
/* calculate weights and prediction
*/
static void calc_pct_wt()
{
	register int j, num, ix;
	register HOMOL_REC *h;
	PREDICT_REC *p;
	double wt, totScor, totPred;

	for (ix=0; ix<atomNum; ix++) {
		p = &pRec[ix];
		num = p->homNum;

/* first sum all the scores of the valid shift data */ 
		totScor = csDbase;
		for (j=0; j<num; j++) {
			h = &p->hom[j];
			if ((h->wtScor > 0) && (h->selectShift >= 0)) {
				totScor += h->wtScor;
			}
		}

/* percent weight for table value */
		p->dswWt = csDbase / totScor;
		totPred = csDbase * p->dswVal;
		
/* find the percentage weight for each homologous shift */
		for (j=0; j<num; j++) {
			h = &p->hom[j];
			wt = h->wtScor;
			h->homWt = 0.0;
			if ((wt > 0) && (h->selectShift >= 0)) {
				totPred += wt * h->shift[h->selectShift];
				h->homWt = wt / totScor;
			}
		}

/* final prediction */
		p->predict = totPred / totScor;
	}

}	/* calc_pct_wt */


/*************************************************************/
/* search integer array for element sNum, return those that match.
*/
static int get_molmat_ix(sArr, sNum, atIx, matches)
register int *sArr, sNum, atIx, *matches;
{
	register int i, count;

	count = 0;
	for (i=0; i<sNum; i++) {
		if (sArr[i] == atIx) {
			matches[count] = i;	
			count++;
		}
	}
	return(count);

}	/* get_molmat_ix */


/*************************************************************/
/* find atoms which have partners 
*/
static int flip_atom(atom)
char *atom;
{
	register int len;

	len = strlen(atom) - 1;

	if (atom[len] == '1') {
		atom[len] = '2';
		return(TRUE);
	}

	if (atom[len] == '2') {
		atom[len] = '1';
		return(TRUE);
	}

	return(FALSE);

}	/* flip_atom */


/*************************************************************/
/* calc standard deviation for our prediction 
*/
static void calc_sdev()
{
	register int i, k, num, obs, count, totNum;
	int wtCount;
	double ySum, ySum2, sDev, tmp;
	double rTerm, dTerm, stddev;
	PREDICT_REC *p;
	HOMOL_REC *h;

	for (i=0; i<atomNum; i++) {
		p = &pRec[i];
		num = p->homNum;

/* compute weighted standard dev of the observed shifts */
		count = totNum = 0;
		ySum = 0.0;
		ySum2 = 0.0;
		for (k=0; k<num; k++) {
			h = &p->hom[k];
			obs = 100 * h->homWt;
			if (obs > 0) {
				count += obs;
				ySum += h->shift[0] * obs;
				ySum2 += (h->shift[0] * h->shift[0]) * obs;
				totNum++;
			}
		}

		stddev = p->dswStd;

/* for snb, it makes more sense to not recalculate stddev */
/*
		if (count > 1) {
			tmp = MAX(0.0, (count*ySum2 - ySum*ySum));
			sDev = sqrt(tmp / (double)(count*(count-1)));
			wtCount = 100 * p->dswWt;
			stddev = ((sDev * count) + (p->dswStd * wtCount)) /
				(count + wtCount);
		}
		else
			stddev = p->dswStd;
*/
	
/* compute random coil standard dev */
		rTerm = p->dswWt * p->dswStd + (1.0 - p->dswWt) * stddev;
		dTerm = p->dswStd / (double)(totNum + 1);
	
		p->predStd = (MAX(rTerm, dTerm)); 
	}
	
}	/* calc_sdev */
