
/*
	file: init.c
	read input data, output files, and parameters for orb
*/

#include "defs.h"

/* global vars */
AMINO_ACID_REC 	aaTable[CODES];
ALIGN_SEQ_REC	seqArr[MAX_SEQ_NUM];
int 			numSeq;
int				alignLen;
int				seqSelect;
FILE 			*outfp, *verbfp, *csifp, *ppmfp;

/* procedure calls */
static void read_parms(), read_input_data(), read_output_files();
static void do_data_menu_help();
static int do_data_menu();
int do_seq_choice_menu();
static FILE *open_output_file();
extern FILE *get_parms_file();
extern void	read_spec_data(), read_dsw_shifts(), read_cs_parms();
extern void	read_wt_homol(), read_wt_struct(), read_wt_mol();
extern void read_one_seq(), read_xalign(), read_shift_files(), exit();


/*************************************************************/
/* main module for initializing orb
*/
void init_orb()
{

	fprintf(stdout, "Parameters file: %s\n", parmsFname);
	read_parms(parmsFname);
	read_input_data();
	read_output_files();

}   /* init_orb */


/*************************************************************/
/* obtain the parameter file and read it.
*/
static void read_parms(fname)
char *fname;
{
	FILE *parmsfp;

	parmsfp = get_parms_file(fname);
	if (parmsfp == NULL)
		exit(1);

	read_spec_data(parmsfp);
	read_dsw_shifts(parmsfp);
	read_wt_homol(parmsfp);
	read_wt_struct(parmsfp);
	read_wt_mol(parmsfp);
	read_cs_parms(parmsfp);
	fclose(parmsfp);
	  
}	/* read_parms */


/*************************************************************/
/* reads all input data from user 
*/
static void read_input_data()
{
	int select;

	select = do_data_menu();

	if (select == 1)
		read_one_seq();

	if (select == 2) {
		read_xalign();
		seqSelect = do_seq_choice_menu();
		read_shift_files();
	}

	fprintf(stdout, "\n");

}	/* read_input_data */


/*************************************************************/
/* process initial data menu selection. What type of data does
** the user have?
*/
static int do_data_menu()
{
	char aStr[MAX_LINE_LEN];
	int num;

	while (TRUE) {

		fprintf(stdout, "\n");
		fprintf(stdout, " Select the number of the appropriate function.\n");
		fprintf(stdout, "\n");
		fprintf(stdout, "   1) Predict shifts given only sequence/structure\n");
		fprintf(stdout, "   2) Predict shifts given homologous sequences\n");
		fprintf(stdout, "   3) help\n");
		fprintf(stdout, "   4) quit\n");
		fprintf(stdout, "\n");

		fprintf(stdout, PROG_PROMPT);
		fscanf(stdin, "%s", aStr);
		num = atoi(aStr);

		if ((num == 1) || (num == 2))
			break;

		else if (num == 3)
			do_data_menu_help();	

		else if (num == 4) {
			fprintf(stdout, "\n Bye!\n\n");
			exit(1);
		}

		else {
			fprintf(stdout, "\nIllegal option entered: %d\n", num);
			fprintf(stdout, "Please try again.\n");
		}
	}

	return(num);

}	/* do_data_menu */


/*************************************************************/
/* help option of the input data menu 
*/
static void do_data_menu_help()
{

	fprintf(stdout, "\n");
	fprintf(stdout, " If you select option #1, orb will try to predict\n");
	fprintf(stdout, " shifts based on Dave Wishart chemical shift tables.\n");
	fprintf(stdout, " Having secondary structure would help orb to make\n");
	fprintf(stdout, " more accurate predictions. Read the documentation\n");
	fprintf(stdout, " on orb to understand the input file format.\n");
	fprintf(stdout, "\n");
	fprintf(stdout, " If you select option #2, orb requires a multiple\n");
	fprintf(stdout, " alignment output file from the xalign program. This\n");
	fprintf(stdout, " data will indicate how your sequence lines up with\n");
	fprintf(stdout, " homologous sequences. For each homologus sequence,\n");
	fprintf(stdout, " orb prompts the user to enter an assigned chemical\n");
	fprintf(stdout, " shift file (if it exists). Make sure you understand\n");
	fprintf(stdout, " the format for these input files before proceeding.\n");
	fprintf(stdout, "\n");

}	/* do_data_menu_help */


/*************************************************************/
/* read in various output files 
*/
static void read_output_files()
{

	fprintf(stdout, "\n");
	fprintf(stdout, " Now we just need to enter some output files.\n");
	fprintf(stdout, "\n");
	fprintf(stdout, " First, specify the file which will contain the\n");
	fprintf(stdout, " final predicted shifts (in PPM format).\n");
	ppmfp = open_output_file(ppmFname);
	fprintf(stdout, "\n\n");

	fprintf(stdout, " Similar to above, specify the file which will\n");
	fprintf(stdout, " contain predicted shifts in an extended PPM format.\n");
	fprintf(stdout, " Additional columns for random coil shifts and\n");
	fprintf(stdout, " confidence factor are printed \n");
	outfp = open_output_file(resultsFname);
	fprintf(stdout, "\n\n");

	fprintf(stdout, " Next, specify the file which will contain the\n");
	fprintf(stdout, " detailed chemical shift parameters and data used\n");
	fprintf(stdout, " in making the predictions above.\n");
	verbfp = open_output_file(verbFname);
	fprintf(stdout, "\n\n");

	fprintf(stdout, " Finally, after orb has made its shift predictions\n");
	fprintf(stdout, " it may be interesting to run the results through\n");
	fprintf(stdout, " the CSI program. Enter the file which will contain\n");
	fprintf(stdout, " the predicted shifts in a format that CSI could use.\n");
	csifp = open_output_file(csiFname);
	fprintf(stdout, "\n\n");

}	/* read_output_files */


/*************************************************************/
/* open an output file 
*/
static FILE *open_output_file(fname)
char *fname;
{
	FILE *fp;
	char sVal[MAX_LINE_LEN];

	fp = fopen(fname, "w");

	if (fp == NULL) {
		fprintf(stdout, PROG_PROMPT);
		fscanf(stdin, "%s", fname);
		fp = fopen(fname, "w");
		if (fp == NULL) {
			sprintf(sVal, "Cannot open output file: %s\n", fname);
			show_err(TRUE, "open_output_file", sVal);
		}
	}

	return(fp);

}	/* open_output_file */


/*************************************************************/
/* Select sequence from alignment for which we predict
** chemical shifts for (interactively from the user).
*/
int do_seq_choice_menu()
{
	int i, select;
	char aStr[MAX_LINE_LEN];

/* display sequence names */

	for (i=0; i<numSeq; i++) {
		fprintf(stdout, "   %d) ID=%s  %s", i+1, seqArr[i].id, seqArr[i].title);
	}
	fprintf(stdout, "   %d) Exit Program\n", numSeq + 1);

	fprintf(stdout, "\n");
	fprintf(stdout, " The sequences shown above have been read from your\n");
	fprintf(stdout, " input file. Enter the number of the sequence for\n");
	fprintf(stdout, " which you wish to predict chemical shifts for.\n");
	fprintf(stdout, "\n");

/* force user to select one of the options */

	while (TRUE) {
		fprintf(stdout, PROG_PROMPT);
		fscanf(stdin, "%s", aStr);
		select = atoi(aStr) - 1;
		if ((select >= 0) && (select <= numSeq))
			break;
		fprintf(stdout, "Invalid sequence num entered: %d\n", select+1);
		fprintf(stdout, "Please try again.\n");
	}

/* user decides to quit program */

	if (select == numSeq) {
		fprintf(stdout, "\n Bye!\n\n");
		exit(0);
	}

/* print user's selection */

	fprintf(stdout, "\n Orb will be predicting chemical shifts for\n");
	fprintf(stdout, "   %s\n", seqArr[select].title);

	return(select);

}	/* do_seq_choice_menu */
