
/*
	file: rseq.c
	purpose: reads single sequence input format file 
*/

#include "defs.h" 

/* procedure calls */
static FILE *read_one_seq_fname();
static int read_one_seq_data(), check_sbuf_line(), check_ubuf_line();
static int read_nmrview_seq_data();
static void read_seq_id_str(), read_seq_title(), set_id_nums();
static void read_seq_data(), store_seq_data();
extern void show_err();
extern int do_seq_choice_menu();


/*************************************************************/
/* reads sequence file (no homologues or ppm data)
*/
void read_one_seq()
{
	register int i, select;
	char sVal[MAX_LINE_LEN];
	FILE *fp;

	numSeq = 0;
	fp = read_one_seq_fname();

	for (i=0; i<MAX_SEQ_NUM; i++) {
		if (TRUE == read_one_seq_data(fp, &seqArr[i]))
			numSeq++;
		else
			break;
	}

	if (numSeq == 0) {
		rewind(fp);
		if (TRUE == read_nmrview_seq_data(fp, &seqArr[i]))
			numSeq++;
	}
	fclose(fp);

	if (numSeq == 0) {
		sprintf(sVal, "Did not find sequence start char %c\n", INP_SEQ_SCH);
		show_err(TRUE, "read_one_seq_data", sVal);
	} else {
		if (numSeq > 1) {
			select = do_seq_choice_menu();
			seqArr[0] = seqArr[select];
		}
	}

	alignLen = seqArr[0].numRes;
	seqArr[0].startIx = 0;
	seqArr[0].endIx = alignLen - 1;
	numSeq = 1;

}   /* read_one_seq */


/*************************************************************/
/* opens sequence data file.
*/
static FILE *read_one_seq_fname()
{
	FILE *fp;
	char sVal[MAX_LINE_LEN], dataFname[MAX_FNAME_LEN];

	fprintf(stdout, "\n");
	fprintf(stdout, " Orb will attempt to predict your sequence based on\n");
	fprintf(stdout, " random coil (or structural) chemical shifts. The\n");
	fprintf(stdout, " input is a sequence file (here is an example file):\n");
	fprintf(stdout, "     >TITLE: my protein\n");
	fprintf(stdout, "     ADFEETVLPAEASPLLKWQNITHRA\n");
	fprintf(stdout, "     CCCBBBBBCCCCCCHHHHHHHHHHC\n");
	fprintf(stdout, "\n");
	fprintf(stdout, " Please enter your sequence data file.\n");
	fprintf(stdout, PROG_PROMPT);

	fscanf(stdin, "%s", dataFname);
	fp = (FILE *)fopen(dataFname, "r");
	if (fp == NULL) {
		sprintf(sVal, "Cannot open dataFile: %s\n", dataFname);
		show_err(TRUE, "read_one_seq_fname", sVal);
	}

	fprintf(stdout, "\n Reading data file: %s\n", dataFname);

	return(fp);

}	/* read_one_seq_fname */


/*************************************************************/
/* Reads nmrview formatted sequence file 
*/
static int read_nmrview_seq_data(fp, rSeq)
FILE *fp;
ALIGN_SEQ_REC *rSeq;
{
	register int ch, count, nd;
	char sVal[MAX_LINE_LEN];
	char aStr[MAX_LINE_LEN];

	nd = fscanf(fp, "%s", aStr);
	count = 0;
	while (nd != EOF) {
		upper_case(aStr);
		ch = pres_3to1(aStr) + 'A';
		if (AA_CVALID(ch)) {
			rSeq->residue[count].pRes = ch;
			rSeq->residue[count].sRes = CH_X;
		} else {
			sprintf(sVal, "Bad 3 letter amino acid code: %s", aStr);
			show_err(TRUE, "read_nmrview_seq_data", sVal);
		}
		rSeq->residue[count].idNum = count + 1;
		count++;
		nd = fscanf(fp, "%s", aStr);
	}
	strcpy(rSeq->id, "SEQ");
	strcpy(rSeq->title, "Input was Nmrview format sequence\n");
	rSeq->numRes = count;

	return(TRUE);

}	/* read_nmrview_seq_data */


/*************************************************************/
/* Reads one sequence from an input file.
*/
static int read_one_seq_data(fp, rSeq)
FILE *fp;
ALIGN_SEQ_REC *rSeq;
{
	register int ch;
	char sVal[MAX_LINE_LEN];
	char pBuf[MAX_ALIGN_LEN], sBuf[MAX_ALIGN_LEN];

/* look for start of sequence */
	ch = getc(fp);
	while ((ch != EOF) && (ch != INP_SEQ_SCH)) {
		ch = getc(fp);
	}

	if (ch == EOF) {
		return(FALSE);
	}

/* read sequence id code */
	read_seq_id_str(fp, rSeq);

/* read sequence title */
	fgets(sVal, MAX_TITLE_LEN-1, fp);
	read_seq_title(sVal, rSeq);

/* read amino acid data */
	read_seq_data(fp, pBuf, sBuf);
	store_seq_data(rSeq, pBuf, sBuf);

/* set residue id numbers if user specified them */
	set_id_nums(sVal, rSeq);
	return(TRUE);

}	/* read_one_seq_data */


/*************************************************************/
/* Reads the residue id numbers
*/
static void set_id_nums(title, rSeq)
char *title;
ALIGN_SEQ_REC *rSeq;
{
	register int i, k, rCount, nd;
	int startId, endId;
	ALIGN_SEQ_REC *seqPtr;
	char *p, sVal[MAX_LINE_LEN];

	seqPtr = &seqArr[0];
	for (k=0; k<seqPtr->numRes; k++) {
		seqPtr->residue[k].idNum = k+1;
	}

	rCount = 0;
	p = strchr(title, OPEN_RES_ID);
	while (p != NULL) {
		nd = sscanf(&p[1], "%d-%d]", &startId, &endId);
		if (nd == 2) {
			for (k=startId; k<=endId; k++) {
				seqPtr->residue[rCount++].idNum = k;
			}
		}
		p = strchr(&p[1], OPEN_RES_ID);
	}

	if (rCount > 0 && rCount != seqPtr->numRes) {
		sprintf(sVal, "*** Warning, Residue Ids specified do not map exactly to sequence");
		fprintf(stdout, "\n%s\n", sVal);
	}

}	/* set_id_nums */


/*************************************************************/
/* Reads and verifies the sequence id code of a sequence.
*/
static void read_seq_id_str(fp, rSeq)
FILE *fp;
ALIGN_SEQ_REC *rSeq;
{
	register int len;
	char sVal[MAX_LINE_LEN], aStr[MAX_LINE_LEN];

	if (EOF == fscanf(fp, "%s", aStr)) {
		sprintf(sVal, "Expecting sequence but EOF found.\n");
		show_err(TRUE, "read_seq_id_str", sVal);
	}

	len = strlen(aStr);
	if (len >= MAX_ID_LEN) {
		sprintf(sVal, "Sequence ID code too long at: %s\n", aStr);
		show_err(TRUE, "read_seq_id_str", sVal);
	}

	strcpy(rSeq->id, aStr);

}	/* read_seq_id_str */


/*************************************************************/
/* Reads and verifies the title of a sequence.
*/
static void read_seq_title(sVal, rSeq)
char *sVal;
ALIGN_SEQ_REC *rSeq;
{
	register int i, len;

	len = strlen(sVal);
	for (i=0; i<len; i++) {
		if ((sVal[i] != ' ') && (sVal[i] != '\t'))
			break;
	}

	strcpy(rSeq->title, &sVal[i]);

	if (sVal[i] == '\n') {
		sprintf(sVal, "Missing sequence title at: %s", rSeq->id);
		show_err(TRUE, "read_seq_title", sVal);
	}

}	/* read_seq_title */


/*************************************************************/
/* Reads one sequence from an input file checking that input
** data is legit.
*/
static void read_seq_data(fp, pBuf, sBuf)
FILE *fp;
char *pBuf, *sBuf;
{
	register int i, ch, len, pLen, sLen;
	char line[MAX_LINE_LEN], newLine[MAX_LINE_LEN];
	char sVal[MAX_LINE_LEN];

	pLen = sLen = 0;
	ch = getc(fp);
	while ((ch != INP_SEQ_SCH) && (ch != EOF)) {

		if ((ch != '\n') && (ch != COMMENT)) {
			line[0] = ch;
			fgets(&line[1], MAX_LINE_LEN-2, fp);
			len = remove_blanks(line, strlen(line), newLine);
			upper_case(newLine);

			if (FALSE == check_ubuf_line(newLine, len)) { 
				if (TRUE == check_sbuf_line(newLine, len))
					for (i=0; i<len; i++)
						sBuf[sLen++] = AA_UC(newLine[i]);
		
				else {
					for (i=0; i<len; i++) {
						ch = AA_UC(newLine[i]);	/* convert to upper case */
						if (AA_CVALID(ch))
							pBuf[pLen++] = ch;
						else {
							sprintf(sVal, "Illegal input line: %s", line);
							show_err(TRUE, "read_seq_data", sVal);
						}
					}
				}
			
				if (pLen > MAX_ALIGN_LEN) { /* too many residues */
					sprintf(sVal, "Sequence has > %d amino acids",
						MAX_ALIGN_LEN);
					show_err(TRUE, "read_seq_data", sVal);
				}
			}
		}
		ch = getc(fp);
	}

	pBuf[pLen] = '\0';
	sBuf[sLen] = '\0';
	ungetc(ch, fp);

}	/* read_seq_data */


/*************************************************************/
/* store the sequence data we read making sure it is ok.
*/
static void store_seq_data(rSeq, pBuf, sBuf)
ALIGN_SEQ_REC *rSeq;
char  *pBuf, *sBuf;
{
	register int i, pLen, sLen;
	char sVal[MAX_LINE_LEN];

	pLen = strlen(pBuf);
	sLen = strlen(sBuf);

	rSeq->numRes = pLen;

	if (pLen < MIN_SEQ_LEN) {
		sprintf(sVal, "Sequence has < %d amino acids.", MIN_SEQ_LEN);
		show_err(TRUE, "store_seq_data", sVal);
	}
	for (i=0; i<pLen; i++) {
		rSeq->residue[i].pRes = pBuf[i];
		rSeq->residue[i].idNum = i+1;
	}

	if (sLen > 0) {
		if (pLen != sLen) {
			sprintf(sVal, "Primary/secondary structures do not line up.");
			show_err(TRUE, "store_seq_data", sVal);
		}
		for (i=0; i<sLen; i++)
			rSeq->residue[i].sRes = sBuf[i];
	}

	else {
		for (i=0; i<pLen; i++)
			rSeq->residue[i].sRes = CH_X;
	}

}	/* store_seq_data */


/*************************************************************/
/* decide if an input line from the sequence file is a 
** secondary structure line.
*/
static int check_sbuf_line(line, len)
char *line;
int len;
{
	register int i;

	for (i=0; i<len; i++) {
		if (AA_SVALID(line[i]) == FALSE)
			return(FALSE);
	}

	return(TRUE);

}	/* check_sbuf_line */


/*************************************************************/
/* decide if an input line from the sequence file is a 
** control alignment line.
*/
static int check_ubuf_line(line, len)
char *line;
int len;
{
	register int i, anchor;
	char ch;

	for (i=0; i<len; i++) {
		ch = line[i];
		anchor = (ch >= '1') && (ch <= '9');
		if ((ch != '-') && (ch != '*') && (anchor == FALSE))
			return(FALSE);
	}

	return(TRUE);

}	/* check_ubuf_line */
